import os
import time
from functools import lru_cache
from typing import List, Tuple
from decimal import Decimal

import tiktoken
from ollama import Client

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.extractor_dataclasses import Base64Image
from parsee.settings import chat_settings


class OllamaModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024 if model.max_output_tokens is None else model.max_output_tokens
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer

        self.client = Client(host='http://localhost:11434' if model.file_path is None else model.file_path)

    def _call_api(self, prompt: str, images: List[Base64Image]) -> str:
        message_content = {
            'role': 'user',
            'content': prompt
        }
        if self.spec.multimodal:
            message_content['images'] = [x.data.encode() for x in images]
        messages = [message_content]
        if self.spec.system_message is not None:
            messages.insert(0, {
            'role': 'system',
            'content': self.spec.system_message
        })
        response = self.client.chat(model=self.spec.internal_name, messages=messages)
        return response["message"]["content"]

    @lru_cache(maxsize=chat_settings.max_cache_size)
    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, num_tokens_input = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        return self._call_api(final_prompt, prompt.available_data if self.spec.multimodal else []), Decimal(0)



from typing import *
from parsee.chat.custom_dataclasses import Message
from decimal import Decimal
from parsee.storage.interfaces import DocumentManager
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.model_loader import get_llm_base_model
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.utils.helper import merge_answer_pieces
from parsee.settings import chat_settings
from tenacity import RetryError
import logging

logger = logging.getLogger(__name__)


def run_chat_with_fallback(message: Message, message_history: List[Message],
             document_manager: DocumentManager, receivers: List[MlModelSpecification],
             most_recent_references_only: bool, show_chunk_index: bool = False,
             single_page_processing_max_images_trigger: Optional[int] = None) -> List[Message]:
    """Runs a chat with a fallback to other models if the first one fails to process the message."""
    logger.info(f"Running chat with fallback")
    logger.debug(f"Message: {message}")
    for spec in receivers:
        try:
            output = run_chat(message, message_history, document_manager, spec,
                     most_recent_references_only, show_chunk_index, single_page_processing_max_images_trigger)
        except RetryError:
            logger.warning(f"RetryError occurred for model {spec.model_id}. Continuing with next model.")
            continue
        logger.debug(f"Output from the model: {output}")
        return output
    logger.warning(f"No model was able to process the message")
    return []


def run_chat(message: Message, message_history: List[Message],
             document_manager: DocumentManager, spec: MlModelSpecification,
             most_recent_references_only: bool, show_chunk_index: bool = False,
             single_page_processing_max_images_trigger: Optional[int] = None) -> List[Message]:
    """Run a chat with a specific model."""
    logger.info(f"Running chat with {spec.model_id}")
    output = []

    model = get_llm_base_model(spec)

    # collect all references if requested
    references = message.references if most_recent_references_only else []
    if not most_recent_references_only:
        added_references = set()
        all_messages = [message] + message_history
        for m in all_messages:
            new_references = [x for x in m.references if x.reference_id() not in added_references]
            for ref in new_references:
                references.append(ref)
                added_references.add(ref.reference_id())

    data = document_manager.load_documents(references, model.spec.multimodal, str(message), model.spec.max_images, chat_settings.min_tokens_for_instructions_and_history, show_chunk_index)
    # for multimodal queries, check if pages have to be processed individually
    process_pages_individually = False
    if type(data) is list:
        # check if pages can be processed one by one
        if single_page_processing_max_images_trigger is not None and len(data) >= single_page_processing_max_images_trigger:
            process_pages_individually = True
    if process_pages_individually:
        answers = []
        cost = Decimal(0)
        for k, img in enumerate(data):
            if k > 0:
                additional_info = f"We are showing you the images contained in the document one by one. The current image is number {k + 1} out of a total of {len(data)}.\n Your last answer ended with the following (make sure that your new answer is valid JSON or similar, as requested; last 500 characters are shown):\n" \
                                  f"{answers[-1][-500:]}"
            else:
                additional_info = f"We are showing you the images contained in the document one by one. The current image is number {k + 1} out of a total of {len(data)}."
            prompt = Prompt(None, f"{message}", additional_info=additional_info, available_data=[img], history=[str(m) for m in message_history])
            current_answer, current_cost = model.make_prompt_request(prompt)
            answers.append(current_answer)
            cost += current_cost
        answer = merge_answer_pieces(answers)
    else:
        prompt = Prompt(None, f"{message}", available_data=data, history=[str(m) for m in message_history])
        answer, cost = model.make_prompt_request(prompt)

    output.append(Message(answer, [], model.spec.model_id, cost=cost))
    cache_info = model.make_prompt_request.cache_info()
    logger.info(f"Chat with {spec.model_id} done. Cache info: {cache_info}")
    return output



import csv
from typing import *

from parsee.extraction.extractor_elements import StructuredTable, ExtractedSource, StructuredRow, StructuredTableCell, DocumentType


def csv_delimiter_simple(csv_content: str) -> Tuple[str, float, int]:

    num_comma = csv_content.count(",")
    num_semicolon = csv_content.count(";")
    divider = 1 if num_semicolon + num_comma == 0 else num_semicolon + num_comma
    return "," if num_comma > num_semicolon else ";", (num_comma if num_comma > num_semicolon else num_semicolon) / divider, num_semicolon + num_comma


def parse_csv(file_path: str, max_rows: Optional[int] = None) -> StructuredTable:

    table = StructuredTable(ExtractedSource(DocumentType.TABULAR, None, None, 0, None), [])
    sniffer = csv.Sniffer()
    all_rows = []
    max_cols = 0
    try:
        f = open(file_path, "r", encoding='utf8')
        data = f.read()
    except UnicodeDecodeError:
        f = open(file_path, "r", encoding='windows-1254')
        data = f.read()
    f.seek(0)
    delimiter_alt, share_alt, total_alt = csv_delimiter_simple(data)
    args = {}
    try:
        args["dialect"] = sniffer.sniff(data, delimiters=[",", ";", "|"])
    except Exception as e:
        args["delimiter"] = delimiter_alt
    if total_alt > 100 and share_alt > 0.6 and "dialect" in args:
        args["dialect"].delimiter = delimiter_alt
    reader = csv.reader(f, **args)
    for k, row in enumerate(reader):
        if max_rows is not None and k+1 > max_rows:
            continue
        if len(row) > max_cols:
            max_cols = len(row)
        all_rows.append(row)
    for row in all_rows:
        r = StructuredRow("body", [])
        table.rows.append(r)
        for col_idx in range(0, max_cols):
            val = row[col_idx] if col_idx <= len(row) - 1 else ""
            r.values.append(StructuredTableCell(str(val), 1, True))
        r.final_values = r.values
    f.close()

    return table



from copy import deepcopy
from decimal import Decimal
from parsee.extraction.extractor_elements import FinalOutputTable, FinalOutputTableColumn, StructuredTable, StructuredRow, ParseeLocation, ExtractedSource, DocumentType, StructuredTableCell, MappingSchema
from parsee.templates.mappings import MappingBucket, AggregationMethod


def FreezeProperty(value):
    cache = deepcopy(value)
    return property(
        lambda self: deepcopy(cache)
    )


def FrozenSpace(**args):
    args = {k: FreezeProperty(v) for k, v in args.items()}
    args['__slots__'] = ()
    cls = type('FrozenSpace', (), args)
    return cls()


__source_var = ExtractedSource(DocumentType.PDF, None, None, 1, None)


samples = FrozenSpace(
    source = __source_var,
    table = FinalOutputTable("test", [
        FinalOutputTableColumn(ParseeLocation("test", 1.0, "test", 1.0, __source_var, []),
                               StructuredTable(__source_var, [
                                   StructuredRow("body",[
                                       StructuredTableCell("Some item"), StructuredTableCell("31234")
                                   ]),
                                   StructuredRow("body", [
                                        StructuredTableCell("Another item"), StructuredTableCell("591")
                                   ]),
                                   StructuredRow("body", [
                                       StructuredTableCell("Third line item"), StructuredTableCell("492")
                                   ]),
                               ]), 0, 0, 0
                               )
    ], "abc"),
    schema = MappingSchema("test", "Some test mapping", "Just for testing", [
        MappingBucket("id1", None, "Some bucket", None, 0, Decimal(1.0), {}, AggregationMethod.SUM),
        MappingBucket("id2", None, "Some other bucket", None, 0, Decimal(1.0), {}, AggregationMethod.SUM),
        MappingBucket("id3", None, "Third bucket", None, 0, Decimal(1.0), {}, AggregationMethod.SUM),
    ])
)


from typing import Optional

import tiktoken
from tiktoken import Encoding
from pydantic_settings import BaseSettings, SettingsConfigDict


class ChatSettings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8")
    max_el_in_memory: int = 10000
    max_images_to_load_per_doc: int = 30
    min_tokens_for_instructions_and_history: int = 500
    encoding: Encoding = tiktoken.get_encoding("cl100k_base")
    max_cache_size: int = 128
    retry_attempts: int = 5
    retry_wait_multiplier: int = 1
    retry_wait_min: int = 2
    retry_wait_max: int = 20
    openai_key: Optional[str] = None
    replicate_key: Optional[str] = None
    together_api_key: Optional[str] = None
    anthropic_api_key: Optional[str] = None


chat_settings = ChatSettings()




from typing import *

from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl, ExtractedSource
from parsee.utils.enums import DocumentType, ElementType
from langchain_core.document_loaders.base import BaseLoader
from langchain_community.document_loaders import PyPDFLoader
from langchain_core.documents.base import Document

# requires to install langchain via extras:
# poetry install -E
# or pip install -e


def documents_to_extracted_elements(documents: List[Document], document_type: DocumentType) -> List[ExtractedEl]:

    # langchain only has text elements
    return [ExtractedEl(ElementType.TEXT, ExtractedSource(document_type, None, None, k, el.metadata), el.page_content) for k, el in enumerate(documents)]


def langchain_loader_to_sdf(loader: BaseLoader, document_type: DocumentType, source_identifier: str) -> StandardDocumentFormat:

    pages = loader.load_and_split()

    elements = documents_to_extracted_elements(pages, document_type)

    return StandardDocumentFormat(document_type, source_identifier, elements, None)


from typing import *

from parsee.storage.vector_stores.interfaces import VectorStore
from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl
from parsee.utils.enums import ElementType


class NoVectors(VectorStore):

    def __init__(self):
        pass

    def find_closest_elements(self, document: StandardDocumentFormat, search_element_title: str, keywords: Optional[str], tables_only: bool = True) -> List[ExtractedEl]:
        return document.elements

    def sort_identifiers_by_relevance(self, source_identifiers: Set[str], search_query: str) -> Set[str]:
        return source_identifiers



from typing import *
from dataclasses import dataclass
from enum import Enum

from parsee.utils.constants import ID_NOT_CONFIGURED


TRUTH_PREFIX = "ANSWER"
FEATURE_PREFIX = "FEATURE"


@dataclass(frozen=True)
class DatasetColumn:
    col_idx: int
    col_name: str
    is_feature: bool


class TransformationType(Enum):
    MIN_MAX = "min_max"
    CATEGORICAL = "categorical"
    WORD_VECTORS = "word_vectors"


@dataclass
class MinMaxTransformation:
    min: Optional[float]
    max: Optional[float]


@dataclass
class CategoricalTransformation:
    categories: List[str]


@dataclass
class WordVectorTransformation:
    tokenizer: any
    max_words: int
    col_index: int
    vocab_size: Optional[int]


@dataclass
class Transformation:
    type: TransformationType
    min_max: Optional[MinMaxTransformation]
    categorical: Optional[CategoricalTransformation]
    word_vectors: Optional[WordVectorTransformation]


@dataclass
class ColumnSettings:
    col: DatasetColumn
    transformation: Optional[Transformation]


class ColumnData:
    settings: ColumnSettings
    _values: List[any]

    def __init__(self, settings: ColumnSettings, values: List[any]):
        self.settings = settings
        self._values = [self.clean_value(x) for x in values]

    def clean_value(self, value: any) -> any:
        if self.settings.transformation is not None:
            if self.settings.transformation.type == TransformationType.MIN_MAX:
                value = float(value)
            if self.settings.transformation.type == TransformationType.CATEGORICAL:
                value = str(value)
            if self.settings.transformation.type == TransformationType.WORD_VECTORS:
                value = str(value)
        else:
            # no transformation -> make float
            value = float(value)
        return value

    def add_value(self, value: any, load_in_batches: bool):
        value = self.clean_value(value)
        # handle transformations, tokenizer is handled separately
        if self.settings.transformation is not None:
            if self.settings.transformation.type == TransformationType.MIN_MAX:
                if self.settings.transformation.min_max.min is None or self.settings.transformation.min_max.min > value:
                    self.settings.transformation.min_max.min = value
                if self.settings.transformation.min_max.max is None or self.settings.transformation.min_max.max < value:
                    self.settings.transformation.min_max.max = value
            if self.settings.transformation.type == TransformationType.CATEGORICAL:
                if value not in self.settings.transformation.categorical.categories:
                    self.settings.transformation.categorical.categories.append(value)

        if not load_in_batches:
            self._values.append(value)

    def add_value_simple(self, value: any):
        value = self.clean_value(value)
        self._values.append(value)

    def reset_values(self):
        self._values = []


@dataclass(frozen=True)
class MappingUniqueIdentifier:
    schema_id: str
    li_identifier: str
    kv_index: int


@dataclass(frozen=True)
class MetaUniqueIdentifier:
    detected_class: str
    col_idx: int
    kv_identifier: str


class BaseDatasetRow:
    _dict_values: Dict[str, any]

    def __init__(self, dict_values: Dict[str, any]):
        self._dict_values = {FEATURE_PREFIX + "_" + key: value for key, value in dict_values.items()}
        # check that keys are unique
        if len(self._dict_values.keys()) != len(set(self._dict_values.keys())):
            raise Exception("ids not unique")
        self.base_key_len = 0

    def column_names(self) -> List[str]:

        feature_keys = sorted(self._dict_values.keys())

        return feature_keys

    def to_list(self) -> List[str]:

        values = []
        for col_name in self.column_names():
            if col_name in self._dict_values.keys():
                col_value = str(self._dict_values[col_name]).replace('\x00', " ") # remove unreadable chars
                values.append(col_value)
        return values

    def assign_truth_values(self, values: Dict[str, any]):

        truth_keys = {TRUTH_PREFIX+"_"+key: value for key, value in values.items()}
        self._dict_values = {**self._dict_values, **truth_keys}

    def get_feature(self, feature_name: str) -> Union[any, None]:
        key = FEATURE_PREFIX+"_"+feature_name
        return self._dict_values[key] if key in self._dict_values else None

    def get_value(self, col_name: str, is_feature: bool) -> Union[any, None]:
        key = (FEATURE_PREFIX if is_feature else TRUTH_PREFIX)+"_"+col_name
        return self._dict_values[key] if key in self._dict_values else None

    def add_missing_columns(self, columns: List[str]):
        for col in columns:
            self._dict_values[col] = ID_NOT_CONFIGURED

    def get_dataset_columns(self) -> List[DatasetColumn]:
        output = []
        for k, col_name in enumerate(self.column_names()):
            if col_name.startswith(FEATURE_PREFIX):
                output.append(DatasetColumn(k+self.base_key_len, col_name[len(FEATURE_PREFIX)+1:], True))
            elif col_name.startswith(TRUTH_PREFIX):
                output.append(DatasetColumn(k + self.base_key_len, col_name[len(TRUTH_PREFIX)+1:], False))
        return output

    def get_column_by_name(self, column_name: str) -> Union[None, DatasetColumn]:

        is_feature = FEATURE_PREFIX+"_"+column_name in self.column_names()
        key = FEATURE_PREFIX+"_"+column_name if is_feature else TRUTH_PREFIX+"_"+column_name

        if key in self._dict_values:
            idx = self.column_names().index(key)
            return DatasetColumn(idx, column_name, is_feature)
        return None


class DatasetRow(BaseDatasetRow):
    source_identifier: str
    template_id: str
    element_identifier: any

    def __init__(self, source_identifier: str, template_id: str, element_identifier: any, dict_values: Dict[str, str]):
        super().__init__(dict_values)
        self.source_identifier = source_identifier
        self.template_id = template_id
        self.element_identifier = element_identifier
        self.base_columns = ["source_identifier", "template_id", "element_identifier"]
        self.base_key_len = len(self.base_columns)

    def __str__(self):
        return str(self._dict_values)

    def __repr__(self):
        return str(self)

    def column_names(self) -> List[str]:

        return self.base_columns + super().column_names()

    def to_list(self) -> List[str]:

        return [self.source_identifier, self.template_id, self.element_identifier] + super().to_list()

